#
# Aura Engine (https://gitlab.servus.at/aura/engine)
#
# Copyright (C) 2017-2020 - The Aura Engine Team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Client classes for scheduling API operations.

    - ApiFetcher: Poll data from Steering and Tank API.
    - ApiResult: Holding the results from the API, including error codes and message.

"""


import logging
import queue
import threading
import traceback
from datetime import datetime
from typing import NamedTuple

import confuse

from aura_engine.base.api import SimpleCachedRestApi, SimpleRestApi
from aura_engine.base.config import AuraConfig
from aura_engine.base.lang import private
from aura_engine.base.utils import SimpleUtil as SU
from aura_engine.scheduling.domain import (
    Episode,
    Playlist,
    PlaylistItem,
    Show,
    Timeslot,
)
from aura_engine.scheduling.utils import M3UPlaylistProcessor
from aura_steering_api.models.timeslot import Timeslot as API_TIMESLOT
from aura_tank_api.models.playlist import Playlist as API_PLAYLIST
from aura_tank_api.models.playlist_entry import PlaylistEntry as API_PLAYLIST_ENTRY

ApiResult = NamedTuple(
    "ApiResult",
    [("timeslots", list), ("code", int), ("message", str), ("exception", Exception)],
)


class ApiFetcher(threading.Thread):
    """
    Get data from AURA REST APIs.

    Retrieve timeslots, playlists and playlist items and other metadata as JSON.
    Calls the API endpoints of Steering and Tank.

    """

    logging: logging.Logger
    config: confuse.Configuration
    queue = None
    has_already_fetched = False
    stop_event = None
    api = None
    m3u_processor = None

    # Config for API Endpoints
    url_api_timeslots = None
    url_api_playlist = None
    tank_headers = None

    def __init__(self):
        """
        Initialize the API Fetcher.
        """
        self.config = AuraConfig.instance.config
        self.logger = logging.getLogger("engine")
        cache_location = self.config.general.cache_dir
        self.api = SimpleCachedRestApi(SimpleRestApi(), cache_location)
        self.url_api_timeslots = self.config.api.steering.calendar
        self.url_api_playlist = self.config.api.tank.playlist
        self.queue = queue.Queue()
        self.stop_event = threading.Event()

        tank_session = self.config.api.tank.session
        tank_secret = self.config.api.tank.secret
        self.tank_headers = {
            "Authorization": f"Bearer {tank_session}:{tank_secret}",
            "content-type": "application/json",
        }

        self.m3u_processor = M3UPlaylistProcessor()
        threading.Thread.__init__(self)

    def run(self):
        """
        Fetch timeslot data from the Steering and Tank API.
        """
        try:
            timeslots = self.fetch_timeslots()

            # If nothing is fetched, return an empty list
            if not timeslots:
                self.queue.put(ApiResult([], 0, "Nothing fetched", None))

            # Release the mutex
            self.queue.put(ApiResult(timeslots, 0, "Success", None))
        except Exception as e:
            # Release the mutex
            self.queue.put(
                ApiResult([], 1, f"Error while fetching: {str(e)}", traceback.format_exc())
            )

        # Terminate the thread
        return

    def fetch(self):
        """
        Retrieve fetched data from the queue.
        """
        return self.queue.get()

    def terminate(self):
        """
        Terminate the thread.
        """
        self.logger.info(SU.yellow("[ApiFetcher] Shutting down..."))
        self.stop_event.set()

    #
    # private
    #

    @private
    def fetch_timeslots(self) -> list[Timeslot]:
        """
        Fetch timeslot data from Steering.

        This method also:
            - Filters invalid and unnecessary timeslots.
            - Remaps any API fields to there local modal representation.

        Returns:
            ([Timeslot]): A list of timeslots

        @private
        """
        api_timeslots = None
        url = self.url_api_timeslots
        self.logger.debug("Fetch timeslots from Steering API...")
        result = self.api.get(url)
        api_timeslots = result.json
        if not api_timeslots:
            return

        # FIXME Steering doesn't provide JSON as per a valid JSON schema.
        # Hence we need to split the array manually for now.
        # @see https://gitlab.servus.at/aura/steering/-/issues/154
        api_timetable: list[API_TIMESLOT] = []
        for entry in api_timeslots:
            api_timeslot = API_TIMESLOT.from_dict(entry)
            if self.is_valid_timeslot(entry):
                api_timetable.append(api_timeslot)

        # Transform API timeslots to engine timeslots
        timeslots: list[Timeslot] = []
        for api_ts in api_timetable:
            show = Show(
                id=api_ts.show_id,
                name=api_ts.show_name,
                type=api_ts.show_type,
                cat=api_ts.show_categories,
                funding_cat=api_ts.show_funding_category,
                music_focus=api_ts.show_music_focus,
            )
            # FIXME There is no defined delimiter for multiple entries of following props yet
            # For now only a list with one entry is assigned.
            episode = Episode(
                title=api_ts.title,
                hosts=[api_ts.show_hosts],
                languages=[api_ts.show_languages],
                topics=[api_ts.show_topics],
                memo="",
            )
            # TODO resolve timeslot referenced by `repetition_id`
            timeslot = Timeslot(
                id=api_ts.id,
                repetition_id=api_ts.repetition_of_id,
                start=api_ts.start.timestamp(),
                end=api_ts.end.timestamp(),
                show=show,
                episode=episode,
            )
            # Fetch playlists for timeslot
            playlist_timeslot = self.fetch_playlist(api_ts.playlist_id)
            playlist_schedule = self.fetch_playlist(api_ts.schedule_default_playlist_id)
            playlist_show = self.fetch_playlist(api_ts.show_default_playlist_id)
            timeslot.set_playlists(playlist_timeslot, playlist_schedule, playlist_show)
            self.expand_item_duration(playlist_timeslot)
            self.expand_item_duration(playlist_schedule)
            self.expand_item_duration(playlist_show)
            timeslots.append(timeslot)

        return timeslots

    @private
    def fetch_playlist(self, playlist_id: int) -> Playlist:
        """
        Fetch a playlist from Tank.

        Args:
            playlist_id (int): The ID of the playlist.

        Returns:
            (Playlist): The requested playlist or `None` when `playlist_id` is missing.

        @private
        """
        if not playlist_id:
            return None

        playlist: Playlist = None
        url = self.url_api_playlist.replace("${ID}", str(playlist_id))
        self.logger.debug("Fetch playlist from Tank API...")
        result = self.api.get(url, headers=self.tank_headers)
        SU.log_json(self.logger, result.json)
        api_playlist = API_PLAYLIST.from_dict(result.json)
        if api_playlist:
            playlist = Playlist(
                api_playlist.id, api_playlist.description, api_playlist.playout_mode != "linear"
            )
            entry: API_PLAYLIST_ENTRY
            for entry in api_playlist.entries:
                metadata = None
                # Files do have a metadata object
                if entry.file and entry.file.metadata:
                    metadata = PlaylistItem.Metadata(
                        artist=entry.file.metadata.artist,
                        album=entry.file.metadata.album,
                        title=entry.file.metadata.title,
                    )

                # Convert nanoseconds to seconds
                # TODO remove after tank is using seconds for duration
                duration: float = None
                if entry.duration:
                    duration = SU.nano_to_seconds(entry.duration)

                item = PlaylistItem(entry.uri, duration, 100, metadata)

                # "Hidden Functionality" to feed engine with M3U playlists
                # via Tank's "Stream" playlist item type
                # See https://gitlab.servus.at/aura/engine/-/issues/53
                #
                # TODO In the future this is to be replaced by a generic music pool feature.
                items = self.m3u_processor.spread(item)
                if len(items) > 1:
                    self.logger.info(f"M3U playlist spread to {len(items)} items")

                for i in items:
                    playlist.add(i)

        return playlist

    @private
    def expand_item_duration(self, playlist: Playlist):
        """
        Expand item to the timeslot gap left.

        If some playlist item doesn't have a duration assigned, its duration is expanded to the
        remaining duration of the playlist (= timeslot duration minus playlist items with
        duration).

        If there is more than one item without duration, such items are removed from the
        playlist.

        Args:
            playlist (Playlist): the playlist with items to expand.

        @private
        """
        if not playlist:
            return

        timeslot_duration: float = playlist.get_timeslot().get_duration()
        playlist_duration: float = 0.0
        items_wo_duration: list(PlaylistItem) = []

        for item in playlist.get_items():
            if not item.duration:
                items_wo_duration.append(item)
            else:
                playlist_duration += item.duration

        if len(items_wo_duration) == 1:
            items_wo_duration[0].duration = timeslot_duration - playlist_duration
            self.logger.info(f"Expand duration for playlist item #{items_wo_duration[0]}")

        elif len(items_wo_duration) > 1:
            # This case should actually never happen, as Tank does not allow more than one item w/o
            # duration anymore
            for item in reversed(items_wo_duration[1:-1]):
                msg = f"Delete playlist item without duration: {item}"
                self.logger.error(SU.red(msg))
                item.remove()

    @private
    def is_valid_timeslot(self, timeslot: API_TIMESLOT) -> bool:
        """
        Check if timeslot is relevant for further processing.

        The result contains all timeslots starting with the one currently playing until the first
        timeslot starting after 24 hours (excluded). Note, this might influence resuming, like in
        case of a crash single timeslots which are longer than 12 hours long. Think e.g. live
        broadcasts.

        Args:
            timeslots (dict): The timeslots to be filtered

        Returns:
            (bool): `True` if the timeslot should be used for further processing

        @private
        """
        if "start" not in timeslot:
            return False

        now = SU.timestamp()
        now_plus_24hours = now + (12 * 60 * 60)

        start_time = self.str_to_timestamp(timeslot["start"])
        end_time = self.str_to_timestamp(timeslot["end"])
        if start_time <= now_plus_24hours and (start_time >= now) or (start_time < now < end_time):
            return True

        return False

    @private
    def str_to_timestamp(self, dt_string: str) -> int:
        """
        Convert a datetime string to a timestamp.

        Args:
            dt_string (str): datetime string.

        Returns:
            int: UNIX epoch timestamp.

        @private
        """
        dt = datetime.strptime(dt_string, "%Y-%m-%dT%H:%M:%S")
        return SU.timestamp(dt)
