#
# Aura Engine (https://gitlab.servus.at/aura/engine)
#
# Copyright (C) 2017-2020 - The Aura Engine Team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
A collection of utilities used for scheduling.
"""


from __future__ import annotations

import logging

import confuse

import aura_engine.scheduling as scheduling
from aura_engine.base.config import AuraConfig
from aura_engine.base.utils import SimpleUtil as SU
from aura_engine.resources import ResourceUtil
from aura_engine.scheduling.domain import Playlist, PlaylistItem, Timeslot


class M3UPlaylistProcessor:
    """
    Render a M3U Playlist as a engine compatible playlist dictionary.
    """

    logging: logging.Logger
    config: confuse.Configuration
    playlist_folder = None

    def __init__(self):
        """
        Initialize.
        """
        self.config = AuraConfig.instance.config
        self.logger = logging.getLogger("engine")
        self.playlist_folder = ResourceUtil.playlist_folder_path()

    def spread(self, item: PlaylistItem) -> list:
        """
        Generate individual playlist items out of any given M3U playlist.

        The expected source format is 'playlist://%FILENAME%'. If that is the case, it reads the
        M3U file is read and builds individual `PlaylistItem` objects based on the M3U contents.

        The evaluation is skipped if no playlist folder is configured. If the M3U file is not
        readable, it fails silently by returning an empty list.

        Args:
            item (PlaylistItem): The potential M3U items to spread.

        Returns:
            [PlaylitItem]: List containing one or more items.
        """
        if not self.playlist_folder:
            return item

        result = []
        m3u_items = None

        # It's a M3U Playlist which needs to be spread
        if item.source and item.source.startswith("playlist://"):
            playlist_name = item.source.split("playlist://")[1]
            self.logger.info(f"Spreading items of M3U playlist '{playlist_name}'")
            m3u_items = self.read_m3u_file(self.playlist_folder + playlist_name)
            result += m3u_items
        # It's an ordinary item to be taken as it is
        else:
            result.append(item)

        return result

    def read_m3u_file(self, source_file) -> list:
        """
        Read items from an M3U file.

        Args:
            source_file (str): Path to the M3U file.

        Returns:
            [PlaylistItem]: One or more playlist items.
        """
        items: PlaylistItem = []
        try:
            with open(source_file, "r") as file:
                lines = file.readlines()

                for i in range(0, len(lines)):
                    if lines[i].startswith("#EXTINF:"):
                        metadata = lines[i].split("#EXTINF:")[1].split(",")
                        source = "file://" + lines[i + 1].strip()
                        duration = int(metadata[0])
                        artist = metadata[1].split(" - ")[0].strip()
                        title = metadata[1].split(" - ")[1].strip()
                        m = PlaylistItem.Metadata(artist, "", title)
                        item = PlaylistItem(source, duration, 100, m)
                        items.append(item)
        except EnvironmentError:
            self.logger.error(f"Skipping unreadable M3U playlist file '{source_file}'")
        return items


class TimetableRenderer:
    """
    Display current and next timeslots in ASCII for maintenance and debugging.
    """

    logger: logging.Logger
    scheduler: scheduling.scheduler.AuraScheduler
    timetable: timetable.TimetableService

    def __init__(self, scheduler: scheduling.scheduler.AuraScheduler):
        self.logger = logging.getLogger("engine")
        self.scheduler = scheduler
        self.timetable = scheduler.get_timetable()

    def get_ascii_timeslots(self) -> str:
        """
        Create a printable version of the current timetable.

        The output contains playlists and items as per timeslot.

        Returns:
            (String): An ASCII representation of the current and next timeslots.

        """
        active_timeslot: Timeslot = self.timetable.get_current_timeslot()

        s = "\n\n   SCHEDULED NOW:"
        s += "\n┌".ljust(102, "-")

        if active_timeslot:
            planned_playlist: Playlist = active_timeslot.get_current_playlist()
            s += f"\n│   Playing timeslot {active_timeslot}"

            if planned_playlist:
                t = planned_playlist.get_type()
                s += f"\n│       └── {t}: {planned_playlist}"

                active_item: PlaylistItem = self.timetable.get_current_item()
                item_num: int = 1

                if active_item:
                    # Display previous playlist items
                    prev_items: [PlaylistItem] = active_item.get_all_prev()
                    for item in prev_items:
                        s += self.build_item_string("\n│         └── ", item_num, item, True)
                        item_num += 1

                    # Display current playlist item
                    active_str = SU.green(f"PLAYING > {active_item}]")
                    s += f"\n│         └── Item {item_num} | {active_str}"
                    item_num += 1

                    # Display next playlist items
                    next_items: [PlaylistItem] = active_item.get_all_next()
                    items = self.preprocess_items(next_items, False)
                    s += self.build_playlist_string(items, item_num)
            else:
                s += "\n│      "
                s += SU.red("No playlist assigned. Station fallback will take over.")
        else:
            s += "\n│   Nothing.           "

        s += "\n└".ljust(102, "-")
        s += "\n   SCHEDULED NEXT:"
        s += "\n┌".ljust(102, "-")

        next_timeslots: [Timeslot] = self.timetable.get_next_timeslots()
        if not next_timeslots:
            s += "\n│   Nothing.         "
        else:
            timeslot: Timeslot
            for timeslot in next_timeslots:
                playlist: Playlist = timeslot.get_current_playlist()
                if playlist:
                    t = SU.cyan(str(playlist.get_type()))
                    s += f"\n│   Queued timeslot {timeslot}         "
                    s += f"\n│      └── {t}: {playlist}"
                    if playlist.get_duration() > timeslot.get_end() - timeslot.get_start():
                        warning = SU.red("↑↑↑ Playlist ends after timeslot!")
                        s += f"\n│          {warning}!              "

                    items = self.preprocess_items(playlist.items, False)
                    s += self.build_playlist_string(items, 1)

        s += "\n└".ljust(102, "-") + "\n\n"
        return s

    def build_playlist_string(self, items: [PlaylistItem], num_start: int) -> str:
        """
        Return a stringified list of items.

        Args:
            items ([PlaylistItem]): The items.
            num_start (int): The number of the first item.

        Returns:
            (str): The resulting string to be displayed.
        """
        s = ""
        is_out_of_timeslot = False

        item: PlaylistItem
        for item in items:
            if (
                item.play_queue_state == PlaylistItem.QueueState.OUT_OF_SCHEDULE
                and not is_out_of_timeslot
            ):
                warning = SU.red("↓↓↓ These 'out of timeslot' items won't be played.")
                s += f"\n│             {warning}"
                is_out_of_timeslot = True

            s += self.build_item_string("\n│         └── ", num_start, item, is_out_of_timeslot)
            num_start += 1

        return s

    def build_item_string(self, prefix: str, num: int, item: PlaylistItem, strike: bool) -> str:
        """
        Return an stringified item.

        Args:
            prefix (str): The prefix.
            num (int): The number of the item in the playlist.
            strike (bool): If the string should be striked through.

        Returns:
            (str): The item as string to be displayed.
        """
        s = ""
        if item.play_queue_state == PlaylistItem.QueueState.CUT:
            warning = SU.red("↓↓↓ This item is going to be cut.")
            s = f"\n│             {warning}"

        if strike:
            item_str = SU.strike(item)
        else:
            item_str = str(item)

        return s + f"{prefix}Item {num} | {item_str}"

    def preprocess_items(self, items: [PlaylistItem], cut_oos: bool) -> [PlaylistItem]:
        """
        Analyse and mark items which are going to be cut or excluded.

        Args:
            items ([PlaylistItem]): The playlist items to be scheduled for playout
            cut_oos (Boolean): If `True` items which are 'out of timeslot' are not returned

        Returns:
            ([PlaylistItem]): The list of processed playlist items
        """
        clean_items = []

        item: PlaylistItem
        for item in items:
            timeslot: Timeslot = item.get_playlist().get_timeslot()
            if item.get_start() >= timeslot.get_end():
                msg = f"Filtered item ({item}) after end-of timeslot ({timeslot}) ... SKIPPED"
                self.logger.debug(msg)
                item.play_queue_state = PlaylistItem.QueueState.OUT_OF_SCHEDULE
            elif item.get_end() > timeslot.get_end():
                item.play_queue_state = PlaylistItem.QueueState.CUT
            else:
                item.play_queue_state = PlaylistItem.QueueState.OKAY

            if not item.play_queue_state == PlaylistItem.QueueState.OUT_OF_SCHEDULE or not cut_oos:
                clean_items.append(item)

        return clean_items
