#
# Aura Engine (https://gitlab.servus.at/aura/engine)
#
# Copyright (C) 2017-now() - The Aura Engine Team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import inspect
import time
import unittest

from aura_engine.base.config import AuraConfig
from aura_engine.events import EngineEventDispatcher, EventBinding
from aura_engine.scheduling.domain import PlaylistItem

#
# Mocks
#


class MockedPlayer:
    """
    Mocked version of engine player.
    """

    def roll(channel, seconds_to_roll: int):
        print(f"called pre-rolling for {channel} seconds on player channel {seconds_to_roll}")
        pass


class MockedScheduler:
    """
    Mocked version of scheduler.
    """

    engine = None

    def __init__(self, engine):
        self.engine = engine

    def boot(self):
        print("calling mocked 'boot' in scheduler")

    def on_ready(self):
        """
        Call when the engine has finished booting and is ready to play.
        """
        print(f"called mocked 'on_ready' in scheduler")


class MockedEngine:
    """
    Mocked version of engine.
    """

    scheduler: MockedScheduler
    player: MockedPlayer

    def __init__(self):
        """
        Init.
        """
        self.scheduler = None
        self.player = MockedPlayer


class MockEventHandler:
    engine: MockedEngine
    call_stack: [] = []

    def __init__(self, engine: MockedEngine) -> None:
        self.engine = engine

    def on_initialized(self, _):
        """
        Call when the engine is initialized, just before it is ready.
        """
        m = inspect.stack()[0][3]
        print(f"called mocked '{m}'")
        self.call_stack.append(m)

    def on_boot(self):
        """
        Call when the engine is starting up.
        """
        m = inspect.stack()[0][3]
        print(f"called mocked '{m}'")
        self.call_stack.append(m)

    def on_ready(self, _):
        """
        Call when the engine has finished booting and is ready to play.
        """
        m = inspect.stack()[0][3]
        print(f"called mocked '{m}'")
        self.call_stack.append(m)

    def on_timeslot_start(self, timeslot):
        """
        Call when a timeslot starts.
        """
        m = inspect.stack()[0][3]
        print(f"called mocked '{m}'")
        self.call_stack.append(m)

    def on_timeslot_end(self, timeslot):
        """
        Call when a timeslot ends.
        """
        m = inspect.stack()[0][3]
        print(f"called mocked '{m}'")
        self.call_stack.append(m)

    def on_play(self, item):
        """
        Call by the engine when some play command to Liquidsoap is issued.
        """
        m = inspect.stack()[0][3]
        print(f"called mocked '{m}'")
        self.call_stack.append(m)

    def on_stop(self, channel):
        """
        Call when the passed channel has stopped playing.
        """
        m = inspect.stack()[0][3]
        print(f"called mocked '{m}'")
        self.call_stack.append(m)

    def on_fallback_active(self, timeslot):
        """
        Call when a fallback is activated for the given timeslot.

        This means there is no proper playlist scheduled.
        """
        m = inspect.stack()[0][3]
        print(f"called mocked '{m}'")
        self.call_stack.append(m)

    def on_queue(self, items: list):
        """
        Call when one or more playlist items have been queued and are currently being pre-loaded.

        Args:
            items (list): List of playlist items.
        """
        m = inspect.stack()[0][3]
        print(f"called mocked '{m}'")
        self.call_stack.append(m)

    def on_sick(self, data):
        """
        Call when the engine is in some unhealthy state.
        """
        m = inspect.stack()[0][3]
        print(f"called mocked '{m}'")
        self.call_stack.append(m)

    def on_resurrect(self, data):
        """
        Call when the engine turned healthy again after being sick.
        """
        m = inspect.stack()[0][3]
        print(f"called mocked '{m}'")
        self.call_stack.append(m)


class TestEventDispatcher(unittest.TestCase):
    """
    Testing the event dispatcher.
    """

    config = None

    def setUp(self):
        self.config = AuraConfig.instance.config

    def test_event_dispatcher_init(self):
        print(self._testMethodName)

        engine = MockedEngine()
        dispatcher = EngineEventDispatcher(engine)
        binding = dispatcher.attach(MockEventHandler)
        self.assertIsInstance(binding, EventBinding)

        binding.subscribe("on_initialized")
        binding.subscribe("on_boot")
        binding.subscribe("on_ready")
        binding.subscribe("on_timeslot_start")
        binding.subscribe("on_timeslot_end")
        binding.subscribe("on_play")
        binding.subscribe("on_stop")
        binding.subscribe("on_fallback_active")
        binding.subscribe("on_queue")
        binding.subscribe("on_sick")
        binding.subscribe("on_resurrect")

        scheduler = MockedScheduler(MockedEngine())
        item = PlaylistItem("source", 2.3, 100, PlaylistItem.Metadata)
        dispatcher.initialize(scheduler)
        dispatcher.on_boot()
        dispatcher.on_ready()
        dispatcher.on_timeslot_start(None)
        dispatcher.on_timeslot_end(None)
        dispatcher.on_play(item)
        dispatcher.on_stop(None)
        dispatcher.on_fallback_active(None)
        dispatcher.on_queue(item)
        dispatcher.on_sick(None)
        dispatcher.on_resurrect(None)

        time.sleep(4)
        self.assertEqual(11, len(MockEventHandler.call_stack))


if __name__ == "__main__":
    unittest.main()
